using System;

namespace PatrullaAguila
{
	/// <summary>
	/// Descripcin breve de MinHeap.
	/// </summary>
	public class MinHeap
	{
		private NodeHeap[] taulaHeap;
		private int[,,] hash;
		private int elements;
		private int maxX, maxY, maxTemps, maxElements;

		// Constructor
		public MinHeap(int maxX, int maxY, int maxTemps)
		{
			this.maxElements = maxX * maxY * maxTemps;
			taulaHeap = new NodeHeap[maxElements];
			hash = new int[maxX, maxY, maxTemps];
			elements = 1;
			this.maxX = maxX;
			this.maxY = maxY;
			this.maxTemps = maxTemps;
		}

		// Retorna el valor mes petit del heap, el cap.
		public NodeHeap obtenirMesPetit() 
		{
			if(elements == 1) 
			{
				throw new Exception("El MinHeap esta buit!");
			}   
			else 
			{
				return taulaHeap[1];
			}   
		}

		// Inserta un nou node en el Min_Heap.
		public void afegir(NodeHeap node) 
		{
			if(elements == maxElements + 1) 
			{
				throw new Exception("El MinHeap esta ple!");
			}
			else 
			{
				int k = elements;
				elements++;
				bool final = false;
				while(!final && k > 1) 
				{
					if(node.salts < taulaHeap[k / 2].salts) 
					{
						hash[taulaHeap[k / 2].pos.x, taulaHeap[k / 2].pos.y, taulaHeap[k / 2].pos.z] = k;
						taulaHeap[k] = taulaHeap[k / 2];
						k = k / 2;
					}
					else 
					{
						final = true;
					}
				}
				hash[node.pos.x, node.pos.y, node.pos.z] = k;      
				taulaHeap[k].pos = node.pos;
				taulaHeap[k].salts = node.salts;
			}   
		}

		// Consulta si esta buit o no.
		public bool heapBuit() 
		{
			return (elements == 1);
		}

		// Esborra el cap i reordena el heap.
		public void esborrar() 
		{
			if(heapBuit()) 
			{
				throw new Exception("El Min_Heap esta buit!");
			}
			else 
			{
				int k = 1;
				bool final = false;
				while ((k * 2 < elements - 1) && !final) 
				{
					int fill = fillMesPetit(k);
					if (taulaHeap[fill].salts < taulaHeap[elements - 1].salts) 
					{
						hash[taulaHeap[fill].pos.x, taulaHeap[fill].pos.y, taulaHeap[fill].pos.z] = k;
						taulaHeap[k] = taulaHeap[fill];
						k = fill;
					}
					else 
					{
						final = true;
					}   
				}
				hash[taulaHeap[elements - 1].pos.x, taulaHeap[elements - 1].pos.y, taulaHeap[elements - 1].pos.z] = k;
				taulaHeap[k] = taulaHeap[elements - 1];
				elements = elements - 1;
			}  
		}

		// Cerca la posicio del ECT indicada per veure
		// en la corresponent posicio del Min_Heap quin valor hi ha emmagatzemat, en
        // aquest cas el valor guardat es el numero de salts respecte l'origen.
		public int consultarValor(Posicio pos) 
		{
			if (pos.x < 0 || pos.x > maxX || pos.y < 0 || pos.y > maxY ||
				pos.z < 0 || pos.z > maxTemps) 
			{
				throw new Exception("Posicio inexistent en el Heap!");
			}
			else 
			{
				int index = hash[pos.x, pos.y, pos.z];
				return taulaHeap[index].salts;
			}   
		}

		// Modifica el valor d'un element concret del heap i reordena el heap segons el nou valor establert.
		public void canviarValor(Posicio pos, int nouValor)
		{
			if (pos.x < 0 || pos.x > maxX || pos.y < 0 || pos.y > maxY ||
				pos.z < 0 || pos.z > maxTemps) 
			{
				throw new Exception("Posicio inexistent en el Heap!");
			}
			else 
			{
				int index = hash[pos.x, pos.y, pos.z];
				int min = 1;
				taulaHeap[index].salts = nouValor;
				if (index == 1) 
				{  // Si l'element modificat es el cap del heap
					bool fi = valorMajorAFills(index, nouValor, out min);
					while(!fi) 
					{    // Mentre no estigui correctament ordenat
						fi = ordenar(ref index, ref min, Sentit.Down, nouValor);  
					}   
				}
				else 
				{  // Si l'element a canviar no es troba al cap del Min_Heap
					bool avall = false;
					if ((index * 2) + 1 < elements) 
					{
						//int val1 = taulaHeap[index * 2].salts;
						//int val2 = taulaHeap[(index * 2) + 1].salts;
						//if((val1 <= nouValor) || (val2 <= nouValor)) 
						if((taulaHeap[index * 2].salts <= nouValor) || (taulaHeap[(index * 2) + 1].salts <= nouValor)) 
						{
							avall = true;
						}
					}   
					bool amunt = false;
					int val1 = taulaHeap[index / 2].salts;
					if(val1 > nouValor) 
					{
						amunt = true;
					}
					if(avall) 
					{ // Si cal intercanviar el node amb els fills perque
						// es mes gran que algun dels fills
						bool fi = valorMajorAFills(index, nouValor, out min);
						while(!fi) 
						{  // Mentre no estigui correctament ordenat
							fi = ordenar(ref index, ref min, Sentit.Down, nouValor);
						}   
					}
					if(amunt) 
					{  // Si cal intercanviar el node amb el pare perque
						// es menor que el pare
						bool fi = true;
						//int val1 = taulaHeap[index / 2].salts;
						//if(val1 > nouValor) 
						if(taulaHeap[index / 2].salts > nouValor) 
						{
							fi = false;
							min = index / 2;
						}
						while(!fi) 
						{  // Mentre no estigui ordenat correctament
							fi = ordenar(ref index, ref min, Sentit.Up, nouValor);
						}   
					}
				}
			}
		}

		// Obte el fill mes petit del pare indicat.
		private int fillMesPetit(int k) 
		{
			int petit;

			if((k * 2 + 1) == (elements - 1)) 
			{  // Si nomes te fill esquerre
				petit = k * 2;
			}
			else 
			{
				if(taulaHeap[k * 2].salts < taulaHeap[k * 2 + 1].salts) 
				{
					petit = k * 2;
				}
				else 
				{
					petit = k * 2 + 1;
				}
			}   
			return petit;
		}

		// Compara el valor especificat amb els fills
		// de la posicio indicada, i obte si es major o no als valors dels fills, i en
		// cas de ser major retorna el fill menor.
		private bool valorMajorAFills(int index, int nouValor, out int min) 
		{

			bool fi = true;
			int val1 = taulaHeap[index * 2].salts;
			int val2 = taulaHeap[(index * 2) + 1].salts;
			if(val1 < nouValor || val2 < nouValor) 
			{
				fi = false;
				if(val1 < val2) 
				{
					min = index * 2;
				}
				else 
				{
					min = (index * 2) + 1;
				}
			} 
			else 
			{
				min = 0;
			}
			return fi;
		}

		// Permet ordenar el heap durant l'execucio d'un canvi de valor en un elment del Heap.
		private bool ordenar(ref int index, ref int min, Sentit sentit, int nouValor) 
		{

			bool fi = false;
			NodeHeap node = taulaHeap[index];
			taulaHeap[index] = taulaHeap[min];
			taulaHeap[min] = node;
			Posicio posK = taulaHeap[index].pos;
			Posicio posJ = taulaHeap[min].pos;
			hash[posK.x, posK.y, posK.z] = index;
			hash[posJ.x, posJ.y, posJ.z] = min;
			index = min;
			if(sentit == Sentit.Down) 
			{
				if(index == elements) 
				{
					fi = true;
				}
				else 
				{
					fi = valorMajorAFills(index, nouValor, out min);
				}
			}
			else 
			{ // sentit == Up
				if(index == 1) 
				{
					fi = true;
				}
				else 
				{
					fi = true;
					int val1 = taulaHeap[index / 2].salts;
					if (val1 > nouValor) 
					{
						fi = false;
						min = index / 2;
					}
				}
			}
			return fi;
		}
	}
}
