using System;
using System.Threading;

namespace PatrullaAguila
{
	/// <summary>
	/// Descripcin breve de Class1.
	/// </summary>
	class Patrulla
	{
		/// <summary>
		/// Punto de entrada principal de la aplicacin.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			new Patrulla();
		}

		private TipusFormacio formacioAnterior;
		private TipusFormacio formacioActual;
		private Formacio formacions;
		private Planificador planificador;
		private int maxX;
		private int maxY;
		private int maxTemps;
		private TipusDijkstra[,,] ect;
		private static Mutex mut = new Mutex();
		private EventManager eventManager;
		private InputController inputController;
		private SDLViewer sdlViewer;

		public Patrulla() 
		{
			maxX = 7;
			maxY = 7;
			maxTemps = 20;
			formacioAnterior = TipusFormacio.Inici;
			formacioActual = TipusFormacio.Inici;

			// ECT
			ect = new TipusDijkstra[maxX, maxY, maxTemps];
			
			// Carreguem fitxer amb formacions
			formacions = new Formacio();
			formacions.inicialitzarFormacions("../../formacions.txt");

			// Planificador
			planificador = new Planificador(maxX, maxY, maxTemps, formacions, ect);

			// SDL
			eventManager = new EventManager();
			eventManager.OnEntityMoveRequestEvent += new EventManager.EntityMoveRequestEventHandler(this.Moure);
			inputController = new InputController(eventManager);
			sdlViewer = new SDLViewer(eventManager);

			// Formacio inicial
			formacioActual = TipusFormacio.Cunya;
			sdlViewer.ShowFormacio(formacions.obtenir_posicions_formacio(formacioActual));
			formacioAnterior = formacioActual;

			// Espera infinita
			inputController.Run();
		}

		public void Moure(object sender, EntityMoveRequestEventArgs e) 
		{
			if (!mut.WaitOne(0, false)) 
			{
				return;
			}
			
			formacioActual = e.Formacio;
			
			// Planificacio i posterior visualitzacio dels moviments 
			// necessaris de la patrulla per passar de formacio_anterior
			// fins a formacio_actual
			int maxTemps = planificador.planificarMoviments(formacioAnterior, formacioActual);
			TipusCoordenada coord;
			TipusCoordenada[] coordenades = new TipusCoordenada[7];
			for(int temps = 0; temps <= maxTemps; temps++) 
			{
				for(int i=0; i<7; i++) 
				{
					posicioAvio(this.ect, temps, i, out coord);
					coordenades[i] = coord;
				}
				Thread.Sleep(750);
				sdlViewer.ShowFormacio(coordenades);
			}

			formacioAnterior = formacioActual;

			mut.ReleaseMutex();
		}

		public bool posicioAvio(TipusDijkstra[,,] ect, int temps, int avio, out TipusCoordenada coord)
		{
			bool trobat = false;
			int i = 0;
			int j = 0;
			while(!trobat && i < 7)		// maxX = 7
			{
				if (ect[i, j, temps].avio == avio) 
				{
					trobat = true;
				}
				else 
				{
					j++;
					if (j == 7)			// maxY = 7
					{
						i++;
						j = 0;
					}   
				}    
			}
		
			coord.x = i;
			coord.y = j;
			return trobat;
		}
	}
}
