using System;
using System.Collections;

namespace PatrullaAguila
{
	public struct TipusDijkstra 
	{
		public int avio;
		public int tractat;
		public int pendent;
		public Posicio nodeAnterior;
	}

	/// <summary>
	/// Descripcin breve de Planificador.
	/// </summary>
	public class Planificador
	{
		private TipusDijkstra[,,] ect;
		private int numMovimentsMajor;
		private Formacio formacions;
		int maxX, maxY, maxTemps;


		public Planificador(int maxX, int maxY, int maxTemps, Formacio formacions_ini, TipusDijkstra[,,] referencia_ect)
		{
			//etc = new TipusDijkstra[maxY, maxTemps];
			this.ect = referencia_ect;
			this.formacions = formacions_ini;
			this.maxX = maxX;
			this.maxY = maxY;
			this.maxTemps = maxTemps;
		}

		// Es crida externament per 
		// planificar les trajectories entre dues formacions d'avions i que finalment
		// deixa en el ECT tots els passos necessaris per realitzar els moviments
		// transitoris entre formacions.
		public int planificarMoviments(TipusFormacio anterior, TipusFormacio actual) 
		{
			TipusCoordenada[] coordsAnteriors = formacions.obtenir_posicions_formacio(anterior);
			TipusCoordenada[] coordsActuals = formacions.obtenir_posicions_formacio(actual);
			inicialitzarETC();
			bool camiInexistent = false;
			int numAvio = 0;
			numMovimentsMajor = 0;
			while(numAvio < 7 && !camiInexistent) // 7 avions
			{
				camiInexistent = dijkstra(coordsAnteriors[numAvio], coordsActuals[numAvio], numAvio);
				numAvio++;
			}
			if(camiInexistent) 
			{
				return -1;
			}
			else 
			{
				return numMovimentsMajor;
			}
		}

		// Inicialitza els valors del ECT abans de aplicar el calcul planificador de canvi d'una formacio a una altra.
		public void inicialitzarETC() 
		{
			for(int i = 0; i < maxX; i++) 
			{
				for(int j = 0; j < maxY; j++) 
				{
					for(int k = 0; k < maxTemps; k++) 
					{
						ect[i, j, k].avio = -1;
						ect[i, j, k].tractat = -1;
						ect[i, j, k].pendent = -1;
					}   
				}   
			}   
		}

		// Obte el nombre minim de moviments per arribar de la posicio inicial a la posicio final.
		private	int calcularSaltsCamiMinim(TipusCoordenada origen, TipusCoordenada desti) 
		{
			int num_salts;
			int diferenciaX = 0;
			int diferenciaY = 0;
			int xini = origen.x;
			int yini = origen.y;   
			int xfi = origen.y;
			if(xini > desti.x) 
			{
				while (xini > desti.x) 
				{
					diferenciaX++;
					xini--;
				}   
			}
			else 
			{
				while (xini < desti.x) 
				{
					diferenciaX++;
					xini++;
				}   
			}
			if(yini > desti.y) 
			{
				while (yini > desti.y) 
				{
					diferenciaY++;
					yini--;
				}   
			}
			else 
			{
				while (yini < desti.y) 
				{
					diferenciaY++;
					yini++;
				}   
			}
			num_salts = diferenciaX + diferenciaY;
			return num_salts;
		}

		// Crea un struct Posicio
		private Posicio assignar_posicio(int x, int y, int z) 
		{
			Posicio pos;
			pos.x = x;
			pos.y = y;
			pos.z = z;
			return pos;
		}

		// A partir d'una posicio determinada del ECT retorna un vector amb tots els possibles successors tenint en compte
		// els limits del ECT i els obstacles existents actualment en el ECT.
		private void calcularSuccessors(Posicio posActual, ArrayList successors) 
		{
			int x = posActual.x;
			int y = posActual.y;
			int temps = posActual.z;
			if((x > 0) && (temps < maxTemps - 1)) 
			{ // Si no es la primera posicio de x
				if((ect[x - 1, y, temps + 1].avio == -1)) 
				{
					if ((ect[x, y, temps + 1].avio == -1) || (ect[x - 1, y, temps].avio == -1)) 
					{
						successors.Add(assignar_posicio(x - 1, y, temps + 1));
					}
					else 
					{
						// Si no hi ha creuament
						if(ect[x + 1, y, temps].avio != ect[x, y, temps + 1].avio) 
						{  // Si la posicio horitzontal anterior esta lliure ) {
							successors.Add(assignar_posicio(x -1, y, temps + 1));     
						}   
					}         
				}
			}
			if((x < maxX - 1) && (temps < maxTemps - 1)) 
			{ // Si no es la ultima posicio de x
				if((ect[x + 1, y, temps + 1].avio == -1)) 
				{
					if ((ect[x, y, temps + 1].avio == -1) || (ect[x + 1, y, temps].avio == -1)) 
					{
						successors.Add(assignar_posicio(x + 1, y, temps + 1));
					}
					else 
					{
						// Si no hi ha creuament
						if(ect[x + 1, y, temps].avio != ect[x, y, temps + 1].avio) 
						{  // Si la posicio horitzontal seguent esta lliure ) {
							successors.Add(assignar_posicio(x + 1, y, temps + 1));     
						}   
					}
				}
			}   
			if((y > 0) && (temps < maxTemps - 1)) 
			{ // Si no es la primera posicio de y
				if((ect[x, y - 1, temps + 1].avio == -1)) 
				{
					if ((ect[x, y, temps + 1].avio == -1) || (ect[x, y - 1, temps].avio == -1)) 
					{
						successors.Add(assignar_posicio(x,y - 1,temps + 1));
					}
					else 
					{
						// Si no hi ha creuament
						if(ect[x, y - 1, temps].avio != ect[x, y, temps + 1].avio) 
						{  // Si la posicio vertical anterior esta lliure ) {
							successors.Add(assignar_posicio(x, y - 1, temps + 1));     
						}   
					}         
				}
			}
			if((y < maxY - 1) && (temps < maxTemps - 1)) 
			{  // Si no es la ultima posicio de y
				if((ect[x, y + 1, temps + 1].avio == -1)) 
				{
					if ((ect[x, y, temps + 1].avio == -1) || (ect[x, y + 1, temps].avio == -1)) 
					{
						successors.Add(assignar_posicio(x, y + 1, temps + 1));
					}
					else 
					{
						// Si no hi ha creuament
						if(ect[x, y + 1, temps].avio != ect[x, y, temps + 1].avio) 
						{  // Si la posicio vertical seguent esta lliure ) {
							successors.Add(assignar_posicio(x, y + 1, temps + 1));
						}   
					}                 
				}
			}
			if(temps < maxTemps - 1) 
			{               // Si no es la ultima posicio de temps
				if(ect[x, y, temps + 1].avio == -1) 
				{  // Si la posicio seguent esta lliure
					successors.Add(assignar_posicio(x, y, temps + 1));
				}
			}
		}

		// Obte a partir de la posicio inicial
		// el graf necessari del ECT i del calcul de sucessors per tal de poder 
		// inicialitzar el Min_Heap per a ser posteriorment tractat.
		private void obtenirGrafIPrepararHeap(MinHeap heap, Stack pila, Posicio pos, int numAvio, ref int numPendents) 
		{
			pila.Push(pos);
			Posicio posIni = pos;
			while(pila.Count > 0) 
			{                  // Mentre la pila no sigui buida
				Posicio posActual = (Posicio)pila.Pop();    // Obtencio del cim de la pila
				if(ect[posActual.x, posActual.y, posActual.z].tractat != numAvio) 
				{
					NodeHeap elemHeap;
					elemHeap.pos = posActual;
					if ((posActual.x == posIni.x) && (posActual.y == posIni.y)
						&& (posActual.z == posIni.z)) 
					{
						elemHeap.salts = 0;
					}
					else 
					{
						elemHeap.salts = maxTemps + 1;
					}
					heap.afegir(elemHeap);
					numPendents = numPendents + 1;
					ArrayList llistaFills = new ArrayList();
					calcularSuccessors(posActual, llistaFills);

					//std::vector<TPosicio>::const_iterator itFills;
					// Iteracio dels successors de la posicio actual, apilant els no tractats
					foreach(Posicio fill in llistaFills)
					{
						if(ect[fill.x, fill.y, fill.z].tractat != numAvio) 
						{ // Si la posicio no s'ha tractat
							pila.Push(fill);
						}
					}   
					ect[posActual.x, posActual.y, posActual.z].tractat = numAvio;
				}
			}
		}

		// Mtode que s'executa una vegada ha estat executat
		// el metode d'obtencio de graf i inicialitzacio de heap, i que cerca en el heap
		// els valors mes petits fins a trobar la posicio final, tot aplicant l'algorisme
		// de Dijkstra.
		private int cercarCamiMinim(MinHeap heap, int numPendents, Posicio posFi, int numAvio) 
		{
			while(numPendents > 0) { 
				NodeHeap infoHeap = heap.obtenirMesPetit();
				Posicio posHeap = infoHeap.pos;
				int saltsActuals = infoHeap.salts;
				heap.esborrar();
				if((posHeap.x == posFi.x) && (posHeap.y == posFi.y) && 
					posHeap.z >= posFi.z) {
					return posHeap.z; // Es retorna el temps en el que arriba a la posicio final.
				}
				else {
					ArrayList llistaFills = new ArrayList();
					calcularSuccessors(posHeap, llistaFills);
					
					// Iteracio dels successors de la posicio actual.
					foreach(Posicio fill in llistaFills) {
						if(ect[fill.x, fill.y, fill.z].pendent != numAvio) { // Si no s'ha examninat previament.
							Posicio posFill = assignar_posicio(fill.x, fill.y, fill.z);
							int saltsFill = heap.consultarValor(posFill);
							if(saltsFill > (saltsActuals + 1)) {
								// Es modifica el valor de aquells en els que el valor
								// del Min_Heap sigui major que els salts acumulats del node + 1.
								heap.canviarValor(posFill, saltsActuals + 1);
								// El successor fa guarda la referencia dela posicio predecessora:
								// el salt anterior en el cami de la posicio inicial a la final.
								ect[fill.x, fill.y, fill.z].nodeAnterior = posHeap;                  
							}
						}
					}
					ect[posHeap.x, posHeap.y, posHeap.z].pendent = numAvio;
					numPendents = numPendents - 1;
				}
			}
			return -1; // En cas de no haver trobat un cami possible es retorna 1
		}

		// Assigna al ECT la sequencia de moviments
		// calculada amb el Dijkstra per moure l'avio actual de l'inici fins al desti.
		private void assignarCamiAect(Posicio inici, Posicio desti, int numAvio) 
		{
			// Moviments de l'avio durant el ECT fins a situar-se en la seva posicio final
			Posicio actual = desti;
			while((actual.x != inici.x) || (actual.y != inici.y) || (actual.z != inici.z)) 
			{
				ect[actual.x, actual.y, actual.z].avio = numAvio;
				actual = ect[actual.x, actual.y, actual.z].nodeAnterior;
			}
			ect[actual.x, actual.y, actual.z].avio = numAvio;   
			// Es completa el ECT mantenint fixa la posicio de l'avio fins a MAX_TEMPS
			int numMoviment = desti.z + 1;
			while(numMoviment < maxTemps) 
			{
				// En cas de que un avio amb un nombre major de moviments que hagi estat
				// previament planificat s'ubiqui en la posicio final de l'actual avio,
				// es realitza anticipadament un moviment d'evasio per esquivar l'avio i
				// posteriorment tornar a la posicio original de finalitzacio.
				if(ect[desti.x, desti.y, numMoviment].avio != -1) 
				{
					ArrayList llistaFills = new ArrayList();
					desti.z = numMoviment - 1;
					calcularSuccessors(desti, llistaFills);

					Posicio posEvasio = (Posicio)llistaFills[0];
					ect[posEvasio.x, posEvasio.y, posEvasio.z].avio = numAvio;
				}
				else 
				{
					ect[desti.x, desti.y, numMoviment].avio = numAvio;
				}   
				numMoviment++;
			}   
		}

		private bool dijkstra(TipusCoordenada coordIni, TipusCoordenada coordFi, int numAvio) 
		{
			Stack pila = new Stack();
			MinHeap heap;
			Posicio posIni, posFi;
			int numPendents = 0;
			int temps = 0;
   
			posIni.x = coordIni.x;
			posIni.y = coordIni.y;
			posIni.z = temps;
			posFi.x = coordFi.x;
			posFi.y = coordFi.y;
			posFi.z = calcularSaltsCamiMinim(coordIni, coordFi);
			heap = new MinHeap(this.maxX, this.maxY, this.maxTemps);
			obtenirGrafIPrepararHeap(heap, pila, posIni, numAvio, ref numPendents);
			int numMoviments = cercarCamiMinim(heap, numPendents, posFi, numAvio);
			if(numMoviments != -1) 
			{
				posFi.z = numMoviments; // Unitat de temps en la que s'assoleix la posicio final
				assignarCamiAect(posIni, posFi, numAvio);
				if(numMoviments > numMovimentsMajor) 
				{
					numMovimentsMajor = numMoviments;
				}
				return false;
			}
			else 
			{
				return true;
			}         
		}
	}
}
